"""
Module with models for calculating melt Fe-Mg partition coefficients between olivine and melt as:

 (Fe\ :sup:`2+` / Mg)\ :sub:`ol` / (Fe\ :sup:`2+` / Mg)\ :sub:`melt`
"""

from MagmaPandas.Kd.Ol_melt.Kd_olmelt_FeMg_calculate import *
from MagmaPandas.Kd.Ol_melt.Kd_olmelt_FeMg_models import Kd_olmelt_FeMg_models_dict
