"""
Module with |CO2|-|H2O| solublity models for silicate melts
"""

from MagmaPandas.volatile_solubility.calculate_volatiles import *
from MagmaPandas.volatile_solubility.isobars import *
from MagmaPandas.volatile_solubility.solubility_baseclass import Solubility_model
